/*
 * This program is free software; you can redistribute it
 * and/or modify it under the terms of the Shared Code
 * License (SCL) as published by www.IntellectualHeaven.com;
 * either version 1.0, or (at your option) any later version.
 *
 * You should have received a copy of the SCL 1.0 license,
 * look for file license.txt, along with this software; if not,
 * then please write to x_pankaj_x@intellectualheaven.com
 * for an electronic copy of the same.
 */

/* ++
History (in reverse order)

x_pankaj_x@intellectualheaven.com   2003-08-18
First version checked in CVS
-- */


/* des.cpp -- DES implementation source file */

/***************************************************************
Important thing to remember
8 bytes of input text in DES is represented as left->right bits
e.g.
0x4e,0x6f,0x77,0x20,0x69,0x73,0x20,0x74 is represented as
0100 1110 0110 1111 0111 0111 0010 0000 and so on

and it is stored in the same way in our bytes and that is why
we reverse the bits

so byte[0] = 0; byte[1] = 1; byte[2] = 0; byte[3] = 0; ....
****************************************************************/


//System Include Files
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <winsock2.h>

//User Include Files
#include "des.h"
#include "global_data.h"


//Constructor
CDes::CDes()
{
	for (int i = 0; i < 16; ++i)
		memset((void *)m_key_sched[i], 0, 6);
	m_input_text[8] = 0;
	m_output_text[8] = 0;
}

//Destructor
CDes::~CDes()
{
}

//Copy a bit from src to dest
void CDes::copy_bit(	unsigned char *dest,		//Destination
						int bit_pos_dest,	//Position(0->7) where bit to be copied
						const unsigned char *src,	//Source
						int bit_pos_src		//Position(1->8) from where bit is to be copied
					)
{
	dest[bit_pos_dest] = src[bit_pos_src-1];
}

//Print data as Hexa-Decimal
void CDes::print_hex(const char *sz_title, const unsigned char *text, int n_length)
{
	printf("\n%s", sz_title);
	int i = 0;
	
	for( i = 0; i < (((n_length-1)/8) + 1); ++i)
	{
		unsigned char ch = 0;
		int n_rest = (n_length - i*8);
		if (n_rest >= 8)
		{
			for(int j = 0; j < 8; ++j)
			{
				ch |= (text[i*8 + j] << (7-j));
			}
			printf("%x ", ch);
		}
		else
		{
			for(int j = 0; j < n_rest; ++j)
			{
				ch |= (text[i*8 + j] << (7-j));
			}
			printf("%x ", ch);
		}
	}
}

//Print data as Binary
void CDes::print_binary(const char *sz_title, const unsigned char *text, int n_length)
{
	printf("\n%s", sz_title);
	for (int i = 0; i < n_length; ++i)
		printf("%1d", text[i]);
}

//Calculates Encrypt Key Schedule
void CDes::calculate_encrypt_key_schedule()
{
	unsigned char c[28], d[28];
	
	for (int i = 0; i < 28; ++i)
		copy_bit((unsigned char *)c, i, (unsigned char *)m_key, g_pc_1c[i]);
	
	for (i = 0; i < 28; ++i)
		copy_bit((unsigned char *)d, i, (unsigned char *)m_key, g_pc_1d[i]);

	for (i = 0; i < 16; ++i)
	{
		calculate_next_lt(c, d, g_l_shift[i]);
		for(int j = 0; j < 48; ++j)
		{
			if (g_pc_2[j] <= 28)
				copy_bit((unsigned char *)m_key_sched[i], j, (unsigned char *)c, g_pc_2[j]);
			else
				copy_bit((unsigned char *)m_key_sched[i], j, (unsigned char *)d, g_pc_2[j] - 28);
		}
	}
}

//Calculate Next Left Shift
void CDes::calculate_next_lt(unsigned char *c, unsigned char *d, int n_bits)
{
	unsigned char tempc[2];
	unsigned char tempd[2];
	if (n_bits == 2)
	{
		tempc[0] = c[0];
		tempc[1] = c[1];
		tempd[0] = d[0];
		tempd[1] = d[1];
		memmove(c, c+2, 26);
		memmove(d, d+2, 26);
		c[26] = tempc[0];
		c[27] = tempc[1];
		d[26] = tempd[0];
		d[27] = tempd[1];
	}
	else
	{
		tempc[0] = c[0];
		tempd[0] = d[0];
		memmove(c, c+1, 27);
		memmove(d, d+1, 27);
		c[27] = tempc[0];
		d[27] = tempd[0];
	}
}

//Calculates Decrypt Key Schedule
void CDes::calculate_decrypt_key_schedule()
{
	unsigned char c[28], d[28];
	
	for (int i = 0; i < 28; ++i)
		copy_bit((unsigned char *)c, i, (unsigned char *)m_key, g_pc_1c[i]);
	
	for (i = 0; i < 28; ++i)
		copy_bit((unsigned char *)d, i, (unsigned char *)m_key, g_pc_1d[i]);

	for (i = 0; i < 16; ++i)
	{
		calculate_next_rt(c, d, g_r_shift[i]);
		for(int j = 0; j < 48; ++j)
		{
			if (g_pc_2[j] <= 28)
				copy_bit((unsigned char *)m_key_sched[i], j, (unsigned char *)c, g_pc_2[j]);
			else
				copy_bit((unsigned char *)m_key_sched[i], j, (unsigned char *)d, g_pc_2[j] - 28);
		}
	}
}

//Calculate Next Right Shift
void CDes::calculate_next_rt(unsigned char *c, unsigned char *d, int n_bits)
{
	unsigned char tempc[2];
	unsigned char tempd[2];
	if (n_bits == 2)
	{
		tempc[0] = c[26];
		tempc[1] = c[27];
		tempd[0] = d[26];
		tempd[1] = d[27];
		memmove(c+2, c, 26);
		memmove(d+2, d, 26);
		c[0] = tempc[0];
		c[1] = tempc[1];
		d[0] = tempd[0];
		d[1] = tempd[1];
	}
	else if (n_bits == 1)
	{
		tempc[0] = c[27];
		tempd[0] = d[27];
		memmove(c+1, c, 27);
		memmove(d+1, d, 27);
		c[0] = tempc[0];
		d[0] = tempd[0];
	}
}

//Calculate final cipher bits
void CDes::calculate_final_bits()
{
	unsigned char left_0[32], right_0[32];
	unsigned char left_1[32], right_1[32];

	int bit_pos = 58;
	for (int i = 0; i < 32; ++i)
	{
		copy_bit((unsigned char *)left_0, i, (unsigned char *)m_input_text, bit_pos);
		bit_pos = (bit_pos - 8) >= 0 ? (bit_pos - 8) : (bit_pos + 58);
	}
	
	bit_pos = 57;
	for (i = 0; i < 32; ++i)
	{
		copy_bit((unsigned char *)right_0, i, (unsigned char *)m_input_text, bit_pos);
		bit_pos = (bit_pos - 8) >= 0 ? (bit_pos - 8) : (bit_pos + 58);
	}

	for (i = 0; i < 16; ++i)
	{
		memcpy(left_1, right_0, 32);

		unsigned char e_right[48]; //expanded right
		for (int j = 0; j < 48; ++j)
			copy_bit((unsigned char *)e_right, j, (unsigned char *)right_0, g_expand[j]);

		for (j = 0; j < 48; ++j)
			e_right[j] = (e_right[j] ^ m_key_sched[i][j]);

		unsigned char temp_res[32];
		for (j = 0; j < 8; j++)
		{
			unsigned char row = 2*(e_right[j*6]) + (e_right[j*6 + 5]);
			unsigned char col = 0;
			col = e_right[j*6 + 1]*8 + e_right[j*6 + 2]*4 + e_right[j*6 + 3]*2 + e_right[j*6 + 4];
			temp_res[j*4 + 0] = (S_Box[j][row][col] & 8)/8;
			temp_res[j*4 + 1] = (S_Box[j][row][col] & 4)/4;
			temp_res[j*4 + 2] = (S_Box[j][row][col] & 2)/2;
			temp_res[j*4 + 3] = S_Box[j][row][col] & 1;
		}

		unsigned char result[32];
		for(j = 0; j < 32; ++j)
		{
			copy_bit((unsigned char *)result, j, (unsigned char *)temp_res, g_permute[j]);
		}

		for (j = 0; j < 32; ++j)
			right_1[j] = (left_0[j] ^ result[j]);

		memcpy(left_0, left_1, 32);
		memcpy(right_0, right_1, 32);
	}

	for(i = 0; i < 64; ++i)
	{
		if (g_ini_permute_inv[i] <= 32)
			copy_bit((unsigned char *)m_output_text, i, (unsigned char *)right_1, g_ini_permute_inv[i]);
		else
			copy_bit((unsigned char *)m_output_text, i, (unsigned char *)left_1, g_ini_permute_inv[i]-32);
	}
}

//Encrypt
void CDes::encrypt(const unsigned char *plain_text,
				   const unsigned char *key,
				   unsigned char *encrypted_text)
{
	int i = 0;

	//Reverse the bits in individual bytes
	//This is required for proper data representation
	for(i = 0; i < 8; ++i)
	{
		for(int j = 0; j < 8; ++j)
		{
			if ((key[i] & (1 << j)) == 0)
				m_key[i*8 + (7-j)] = 0;
			else
				m_key[i*8 + (7-j)] = 1;

			if ((plain_text[i] & (1 << j)) == 0)
				m_input_text[i*8 + (7-j)] = 0;
			else
				m_input_text[i*8 + (7-j)] = 1;
		}
	}
	calculate_encrypt_key_schedule();
	calculate_final_bits();

	//Reverse the bits in individual bytes
	//This is required for proper data representation
	for(i = 0; i < 8; ++i)
	{
		encrypted_text[i] = 0;
		for(int j = 0; j < 8; ++j)
		{
			encrypted_text[i] |= (m_output_text[i*8 + j] << (7-j));
		}
	}

	print_hex("Cipher Text: ", m_output_text, 64);
}
	
//Decrypt
void CDes::decrypt(const unsigned char *encrypted_text,
				   const unsigned char *key,
				   unsigned char *plain_text)
{
	int i = 0;

	//Reverse the bits in individual bytes
	//This is required for proper data representation
	for(i = 0; i < 8; ++i)
	{
		for(int j = 0; j < 8; ++j)
		{
			if ((key[i] & (1 << j)) == 0)
				m_key[i*8 + (7-j)] = 0;
			else
				m_key[i*8 + (7-j)] = 1;

			
			if ((encrypted_text[i] & (1 << j)) == 0)
				m_input_text[i*8 + (7-j)] = 0;
			else
				m_input_text[i*8 + (7-j)] = 1;
			
			
		}
	}

	calculate_decrypt_key_schedule();
	calculate_final_bits();

	//Reverse the bits in individual bytes
	//This is required for proper data representation
	for(i = 0; i < 8; ++i)
	{
		plain_text[i] = 0;
		for(int j = 0; j < 8; ++j)
		{
			plain_text[i] |= (m_output_text[i*8 + j] << (7-j));
		}
	}

	print_hex("Plain Text: ", m_output_text, 64);
}
